
#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

extern "C"{
void gpu_matrix_transpose( uint8_t * m1, uint8_t * o );
}

__global__ void matrix_transpose( uint8_t * m1, uint8_t * o ){
    
    int x, y;

    x = threadIdx.x + blockIdx.x * blockDim.x;
    y = threadIdx.y + blockIdx.y * blockDim.y;


    o[x + 8 * y] = m1[y + x * 8];

    return;
}

void gpu_matrix_transpose(uint8_t * src1, uint8_t * out){

    int i, j;
    uint8_t * m1, *o;
    
    dim3 blockDim(8, 8, 1);
    dim3 gridDim(1, 1, 1);

    /* Allocate zero-copy memory */
    cudaMallocHost((void**) &m1, 8 * 8 * sizeof(uint8_t), cudaHostAllocMapped);
    cudaMallocHost((void**) &o, 8 * 8 * sizeof(uint8_t), cudaHostAllocMapped);

    for(i = 0; i < 8; i++){
        for(j = 0; j < 8; j++){
            m1[i + j * 8] = src1[i + j * 8];
        }
    }

    /* We don't need to specify shared memory or stream */
    matrix_transpose<<<blockDim, gridDim>>>(m1, o);

    /* Sync on stream 0, which is everything */
    cudaDeviceSynchronize();

    for(i = 0; i < 8; i++){
        for(j = 0; j < 8; j++){
            out[i + j * 8] = o[i + j * 8];
        }
    }

    cudaFreeHost((void*) m1);
    cudaFreeHost((void*) o);

    return;
}


