
#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>

typedef float my_type;

__global__ void matrix_multiply( my_type * m1, my_type * m2, my_type * o ){
    
    int x, y, off;

#if 1
    __shared__ my_type divider;
    if(threadIdx.x == 0 && threadIdx.y == 0){
        divider = (float) 2.0;
    }
    __syncthreads(); // Wait for thread(0, 0) to finish update shared var
#endif

    x = threadIdx.x + blockIdx.x * blockDim.x;
    y = threadIdx.y + blockIdx.y * blockDim.y;

    off = x + y * blockDim.x;

    o[off] = m1[off] * m2[off] * ((float) 0.2) / divider;

    return;
}

#define X_SZ 1024
#define Y_SZ 1024
int main(void){

    int i, j, off;
    
    my_type * m1;
    my_type * m2;
    my_type * o;

    dim3 blockDim(64, 64, 1);
    dim3 gridDim(X_SZ/64, Y_SZ/64, 1);

    /* Allocate zero-copy memory */
    cudaMallocHost((void**) &m1, X_SZ * Y_SZ * sizeof(my_type), cudaHostAllocMapped);
    cudaMallocHost((void**) &m2, X_SZ * Y_SZ * sizeof(my_type), cudaHostAllocMapped);
    cudaMallocHost((void**) &o, X_SZ * Y_SZ * sizeof(my_type), cudaHostAllocMapped);

    for(i = 0; i < X_SZ; i++){
        for(j = 0; j < Y_SZ; j++){
            off = j + i * X_SZ;
            m1[off] = i + j;
            m2[off] = i + j;
            o[off] = 0.0;
        }
    }

    /* We don't need to specify shared memory or stream */
    matrix_multiply<<<blockDim, gridDim>>>(m1, m2, o);

    /* Sync on stream 0, which is everything */
    cudaDeviceSynchronize();

    printf("Element 0 %f\n", o[0]); // Avoid optimisation

    cudaFreeHost((void*) m1);
    cudaFreeHost((void*) m2);
    cudaFreeHost((void*) o);

    return 0;
}


